from reactivex import operators as ops
from typing import Callable

from reactivex import Observable


def lambda_operator(func) -> Callable[[Observable], Observable]:
	return lambda stream: stream.pipe(
		ops.map(func),
		ops.filter(lambda x: x is not None)
	)



from pybeamline.algorithms.discovery.heuristics_miner_lossy_counting import heuristics_miner_lossy_counting
from pybeamline.algorithms.discovery.heuristics_miner_lossy_counting_budget import heuristics_miner_lossy_counting_budget
from pybeamline.algorithms.discovery.dfg_miner import simple_dfg_miner



from pybeamline.bevent import BEvent
from typing import Iterable, Optional
from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable


def string_test_source(iterable: Iterable[str], scheduler: Optional[abc.SchedulerBase] = None) -> Observable[BEvent]:
    def subscribe(
            observer: abc.ObserverBase[BEvent], scheduler_: Optional[abc.SchedulerBase] = None
    ) -> abc.DisposableBase:
        trace_id = 1
        for trace in iterable:
            for event in trace:
                observer.on_next(
                    BEvent(event, "case_" + str(trace_id), "Process"))
            trace_id += 1
        observer.on_completed()
        return CompositeDisposable()

    return Observable(subscribe)



from reactivex import just, empty
from reactivex import operators as ops
from reactivex import Observable
from pybeamline.bevent import BEvent
from typing import Callable, Tuple


def infinite_size_directly_follows_mapper() -> Callable[[Observable[BEvent]], Observable[Tuple[str, str]]]:
    map_cases = {}

    def extractor(event: BEvent) -> Observable[Tuple[str, str]]:
        to_ret = None
        if event.get_trace_name() in map_cases.keys():
            to_ret = (map_cases[event.get_trace_name()], event.get_event_name())
        map_cases[event.get_trace_name()] = event.get_event_name()
        if to_ret is None:
            return empty()
        else:
            return just(to_ret)
    return ops.flat_map(extractor)



from typing import List, Callable
from reactivex import Observable, empty, just
from reactivex import operators as ops
from pybeamline.bevent import BEvent
from pandas import DataFrame


def list_to_log(events: List[BEvent]) -> DataFrame:
    list_of_events = []
    for e in events:
        data_attributes = e.event_attributes
        data_attributes.update({"case:" + n: e.trace_attributes[n] for n in e.trace_attributes})
        data_attributes.update({"process:" + n: e.process_attributes[n] for n in e.process_attributes})
        list_of_events.append(data_attributes)
    log = DataFrame(list_of_events)
    return log


def sliding_window_to_log() -> Callable[[Observable[Observable[BEvent]]], Observable[DataFrame]]:
    def o2l(obs: Observable[BEvent]) -> Observable[DataFrame]:
        return obs.pipe(
            ops.to_iterable(),
            ops.flat_map(lambda x: empty() if not x else just(list_to_log(x)))
        )

    return ops.flat_map(o2l)



from typing import Callable
from pm4py.algo.discovery.heuristics.variants.classic import calculate as compute_dfg
from pm4py.objects.heuristics_net.obj import HeuristicsNet
from reactivex import Observable
from reactivex import just, empty
from reactivex import operators as ops
from pybeamline.bevent import BEvent


def heuristics_miner_lossy_counting_budget(
        model_update_frequency=10,
        budget=100,
        dependency_threshold=0.5,
        and_threshold=0.8) -> Callable[[Observable[BEvent]], Observable[HeuristicsNet]]:
    hm = HeuristicsMinerLossyCountingBudget(budget=budget, dependency_threshold=dependency_threshold,
                                            and_threshold=and_threshold)

    def miner(event: BEvent) -> Observable[HeuristicsNet]:
        hm.ingest_event(event)
        if hm.observed_events() % model_update_frequency == 0:
            return just(hm.get_model())
        else:
            return empty()

    return ops.flat_map(miner)


# Class originally developed by Magnus Frederiksen as part of his BSc project at DTU entitled
# "Development of Process Mining and Complex Event Processing using Python"
class HeuristicsMinerLossyCountingBudget:
    def __init__(self, budget=10, dependency_threshold=0, and_threshold=0.8):
        self.__budget = int(budget)  # max length of stored events and relations
        self.__minimum_dependency_threshold = dependency_threshold
        self.__and_threshold = and_threshold

        self.__D_C = dict()  # set of event
        self.__D_R = dict()  # set of relations
        self.__observed_events = 1
        self.__current_bucket = 0

    def ingest_event(self, event):

        # for budget lossy counting, if relation or caseID already exists, the memory will just be replaced and not
        # expanded so we only clean up when the caseID or relation doesnt exist, which results in creating a new one
        if event.get_trace_name() in self.__D_C:  # if caseID already exist
            lastEvent = self.__D_C[event.get_trace_name()]  # localy save former event
            del self.__D_C[event.get_trace_name()]  # replace caseID's former event with new event
            self.__D_C[event.get_trace_name()] = [event.get_event_name(), lastEvent[1] + 1, lastEvent[2],
                                                  event.get_event_time()]

            r_N = (lastEvent[0], event.get_event_name())  # save relation localy

            if r_N in self.__D_R:  # if relation exists in set
                lastRelation = self.__D_R[r_N]  # localy save former relation
                del self.__D_R[r_N]  # replace relation

                diff = (event.get_event_time() - lastEvent[3]) - lastRelation[
                    2]  # incase the user wants the time for each relation
                newTime = lastRelation[2] + (diff / (lastRelation[0] + 1))

                self.__D_R[r_N] = [lastRelation[0] + 1, lastRelation[1], newTime]

            else:  # the relation doesent exist, create it
                while len(self.__D_R) + len(
                        self.__D_C) >= self.__budget:  # if budget is reached when adding a new key + iten
                    self.__bucket_cleaning()  # bucket cleaning time
                self.__D_R[r_N] = (1, self.__current_bucket, event.get_event_time() - lastEvent[3])

        else:  # caseID doesnt exist, create it
            while ((len(self.__D_R) + len(
                    self.__D_C)) >= self.__budget):  # if budget is reached when adding a new key + iten
                self.__bucket_cleaning()  # bucket cleaning time
            self.__D_C[event.get_trace_name()] = (
            event.get_event_name(), 1, self.__current_bucket, event.get_event_time())

        # clean up
        self.__observed_events += 1

    def __bucket_cleaning(self):
        self.__current_bucket += 1  # increment bucket to clean all items not within the new bucket number

        D_Ctobedel = []  # the 2 lists are needed to avoid messing with the comming loops
        D_Rtobedel = []

        for caseID, (eventName, frequency, bucket, time) in self.__D_C.items():  # for all caseIDs' occured
            if frequency + bucket <= self.__current_bucket:  # if not above the bucket threshold on all events occured
                D_Ctobedel.append(caseID)

        for caseID in D_Ctobedel:  # deleted the event
            del self.__D_C[caseID]

        for relation, (frequency, bucket, time) in self.__D_R.items():  # for all relations occured
            if frequency + bucket <= self.__current_bucket:  # if not above the bucket threshold of all relations
                D_Rtobedel.append(relation)

        for relation in D_Rtobedel:  # delete the relations
            del self.__D_R[relation]

    def get_model(self):
        dfg = dict()
        for (A, B), (frequency, bucket, time) in self.__D_R.items():
            dfg[(A, B)] = frequency
        hm = HeuristicsNet(dfg)
        return compute_dfg(hm, dependency_thresh=self.__minimum_dependency_threshold,
                           and_measure_thresh=self.__and_threshold)

    def observed_events(self):  # returns map of events observed
        return self.__observed_events - 1



from pybeamline.filters.filters import \
    retains_activity_filter, excludes_activity_filter, \
    retains_on_event_attribute_equal_filter, excludes_on_event_attribute_equal_filter, \
    retains_on_trace_attribute_equal_filter, excludes_on_trace_attribute_equal_filter



import json
import reactivex
import requests
import hashlib
from pyais.stream import TCPConnection
from pybeamline.bevent import BEvent
from functools import partial


def wikimedia_source():
    return reactivex.create(_fetch_wikimedia_stream)


def ais_source(host='153.44.253.27', port=5631):
    def fetch_with_params(observer, _):
        return _fetch_ais_source(observer, host, port)
    return reactivex.create(fetch_with_params)


def _fetch_wikimedia_stream(observer, _):
    url = "https://stream.wikimedia.org/v2/stream/recentchange"
    headers = {"Accept": "application/json"}

    with requests.get(url, headers=headers, stream=True) as response:
        if response.status_code == 200:
            for line in response.iter_lines():
                if line:
                    try:
                        data = json.loads(line.decode("utf-8"))
                        e = BEvent(data["type"], hashlib.md5(data["title_url"].encode()).hexdigest(), data["wiki"])
                        attributes_to_copy = ["title", "user", "bot", "comment", "server_url", "server_name", "namespace", "revision", "length", "timestamp", "wiki"]
                        for k in attributes_to_copy:
                            if k in data:
                                e.event_attributes[k] = data[k]
                        observer.on_next(e)
                    except json.JSONDecodeError as e:
                        observer.on_error(e)
        else:
            observer.on_error(Exception(f"Error: {response.status_code}"))

    observer.on_completed()


def _fetch_ais_source(observer, host, port):
    for msg in TCPConnection(host, port=port):
        decoded_message = msg.decode()
        ais_content = decoded_message
        if hasattr(ais_content, "status"):
            e = BEvent(ais_content.status.name, ais_content.mmsi, "AIS")
            e.event_attributes["mmsi"] = ais_content.mmsi
            e.event_attributes["speed"] = ais_content.speed
            e.event_attributes["accuracy"] = ais_content.accuracy
            e.event_attributes["lon"] = ais_content.lon
            e.event_attributes["lat"] = ais_content.lat
            e.event_attributes["course"] = ais_content.course
            e.event_attributes["heading"] = ais_content.heading
            e.event_attributes["second"] = ais_content.second
            observer.on_next(e)

    observer.on_completed()



import math
from pm4py.algo.discovery.heuristics.variants.classic import calculate as compute_dfg

from pm4py.objects.heuristics_net.obj import HeuristicsNet
from reactivex import just, empty
from reactivex import operators as ops
from reactivex import Observable
from pybeamline.bevent import BEvent
from typing import Callable


def heuristics_miner_lossy_counting(
        model_update_frequency=10,
        max_approx_error=0.001,
        dependency_threshold=0.5,
        and_threshold=0.8) -> Callable[[Observable[BEvent]], Observable[HeuristicsNet]]:
    hm = HeuristicsMinerLossyCounting(max_approx_error=max_approx_error, dependency_threshold=dependency_threshold, and_threshold=and_threshold)

    def miner(event: BEvent) -> Observable[HeuristicsNet]:
        hm.ingest_event(event)
        if hm.observed_events() % model_update_frequency == 0:
            return just(hm.get_model())
        else:
            return empty()

    return ops.flat_map(miner)

# Class originally developed by Magnus Frederiksen as part of his BSc project at DTU entitled
# "Development of Process Mining and Complex Event Processing using Python"
class HeuristicsMinerLossyCounting:
    def __init__(self, max_approx_error=0.1, dependency_threshold=0, and_threshold=0.8):
        self.__minimum_dependency_threshold = dependency_threshold  # set dependency threshold to be added to the model
        self.__and_threshold = and_threshold  # set the "and" threshold for when 2 edges leave a node on model

        self.__D_C = dict()  # set of event
        self.__D_R = dict()  # set of relations
        self.__observed_events = 1
        self.__bucket_width = int(math.ceil(1 / max_approx_error))  # set bucket width
        self.__modelRefreshRate = self.__bucket_width  # default model refreshrate

    def ingest_event(self, event):
        current_bucket = int(math.ceil(self.__observed_events / self.__bucket_width))  # calculated bucket
        if event.get_trace_name() in self.__D_C:  # if caseID already exist
            last_event = self.__D_C[event.get_trace_name()]  # localy save former event

            del self.__D_C[event.get_trace_name()]  # replace caseID's former event with new event
            self.__D_C[event.get_trace_name()] = [event.get_event_name(), last_event[1] + 1, last_event[2], event.get_event_time()]

            r_N = (last_event[0], event.get_event_name())  # save relation localy

            if r_N in self.__D_R:  # if relation exists in set
                last_relation = self.__D_R[r_N]  # localy save former relation
                del self.__D_R[r_N]  # replace relation

                diff = (event.get_event_time() - last_event[3]) - last_relation[2]  # calculates the new average relation time
                new_time = last_relation[2] + (diff / (last_relation[0] + 1))

                self.__D_R[r_N] = [last_relation[0] + 1, last_relation[1], new_time]
            else:  # the relation doesent exist, create it
                self.__D_R[r_N] = (1, current_bucket - 1, event.get_event_time() - last_event[3])

        else:  # caseID doesn't exist, create it
            self.__D_C[event.get_trace_name()] = (event.get_event_name(), 1, current_bucket - 1, event.get_event_time())

        # bucket cleaning time
        if self.__observed_events % self.__bucket_width == 0.0:
            # the 2 lists are needed to avoid messing with the coming loops
            D_C_del = []
            D_R_del = []

            for caseID, (eventName, frequency, bucket, time) in self.__D_C.items():
                if frequency + bucket <= current_bucket:  # if not above the error threshold on all events occured
                    D_C_del.append(caseID)

            for caseID in D_C_del:  # deleted the event
                del self.__D_C[caseID]

            for relation, (frequency, bucket, time) in self.__D_R.items():
                if frequency + bucket <= current_bucket:  # if not above the error threshold of all relations
                    D_R_del.append(relation)

            for relation in D_R_del:  # delete the relations
                del self.__D_R[relation]

        self.__observed_events += 1

    def get_model(self):
        dfg = dict()
        for (A, B), (frequency, bucket, time) in self.__D_R.items():
            dfg[(A, B)] = frequency
        hm = HeuristicsNet(dfg)
        return compute_dfg(hm, dependency_thresh=self.__minimum_dependency_threshold, and_measure_thresh=self.__and_threshold)

    def observed_events(self):  # return the total number of events observed
        return self.__observed_events



from typing import Callable, Iterable
from pybeamline.bevent import BEvent
from reactivex import Observable
from reactivex import operators as ops


def retains_activity_filter(activity_names: Iterable[str]) -> Callable[[Observable[BEvent]], Observable[BEvent]]:
    return ops.filter(lambda x: x.get_event_name() in activity_names)


def excludes_activity_filter(activity_names: Iterable[str]) -> Callable[[Observable[BEvent]], Observable[BEvent]]:
    return ops.filter(lambda x: x.get_event_name() not in activity_names)


def retains_on_event_attribute_equal_filter(
        attribute_name: str, attribute_values: Iterable
) -> Callable[[Observable[BEvent]], Observable[BEvent]]:
    return ops.filter(
        lambda x: attribute_name in x.event_attributes.keys()
                  and x.event_attributes[attribute_name] in attribute_values)


def excludes_on_event_attribute_equal_filter(
        attribute_name: str, attribute_values: Iterable
) -> Callable[[Observable[BEvent]], Observable[BEvent]]:
    return ops.filter(
        lambda x: attribute_name not in x.event_attributes.keys()
                  or x.event_attributes[attribute_name] not in attribute_values)


def retains_on_trace_attribute_equal_filter(
        attribute_name: str, attribute_values: Iterable
) -> Callable[[Observable[BEvent]], Observable[BEvent]]:
    return ops.filter(
        lambda x: attribute_name in x.trace_attributes.keys()
                  and x.trace_attributes[attribute_name] in attribute_values
    )


def excludes_on_trace_attribute_equal_filter(
        attribute_name: str, attribute_values: Iterable
) -> Callable[[Observable[BEvent]], Observable[BEvent]]:
    return ops.filter(
        lambda x: attribute_name not in x.trace_attributes.keys()
                  or x.trace_attributes[attribute_name] not in attribute_values
    )



import sys
from collections import defaultdict
from typing import Callable
from reactivex import empty, just, Observable
from reactivex import operators as ops
from pybeamline.bevent import BEvent
from pybeamline.sources import xes_log_source_from_file


def behavioral_conformance(model) -> Callable[[Observable[BEvent]], Observable[tuple]]:
    bc = BehavioralConformance(M=model)

    def conformance_checker(event: BEvent) -> Observable[tuple]:
        bc.ingest_event(event)
        case_id = event.get_trace_name()
        if case_id in bc.get_conformance():
            return just((bc.get_conformance(case_id), bc.get_confidence(case_id), bc.get_completeness(case_id)))
        else:
            return empty()

    return ops.flat_map(conformance_checker)


def mine_behavioral_model_from_stream(source) -> tuple:
    bmb = BehavioralModelBuilder()
    source.subscribe(
        on_next=lambda i:
        {
            bmb.ingest_event(i),
        },
        on_error=lambda e, l: print("Error Occurred: {0} to {1}".format(e, l)),
    )
    bmb.end_xes_to_model()

    return bmb.get_model()


# Class originally developed by Magnus Frederiksen as part of his BSc project at DTU entitled
# "Development of Process Mining and Complex Event Processing using Python"
class BehavioralConformance:
    def __init__(self, M=([], defaultdict(list), defaultdict(list))):
        # the input is expecting to be the model
        self.set_model(M)  # calls method to set, B, P, and F
        self.__trace_last_event = dict()  # recalls last event for trace to find relation
        self.__conformance = dict()  # saves a traces' conformance
        self.__completeness = dict()  # saves a traces' completeness
        self.__confidence = dict()  # saves a traces' confidence
        self.__obs = defaultdict(list)  # saves all distinct relations in a trace that has ocurred
        self.__inc = dict()  # saves amount of incorrect relations acording to the reference model of a trace

    def ingest_event(self, event):
        case_id = event.get_trace_name()  # easier reference to caseID
        event_name = event.get_event_name()  # easier trefermce to eventNa,e

        if case_id not in self.__trace_last_event:  # if this is first time caseID appears
            self.__trace_last_event[case_id] = event_name  # save current event

            # self.__obs[caseID] = []
            self.__inc[case_id] = 0  # set amount of incorrect relations for that CaseId to 0

        else:  # if the caseID has been seen before

            new_pattern = (self.__trace_last_event[case_id], event_name)  # locally save the relation
            # Step 1: update internal data structures
            if new_pattern in self.__B:  # if the relation is in the approved relation list
                if new_pattern not in self.__obs[case_id]:  # and that relation has not occurred for that caseID before
                    self.__obs[case_id].append(new_pattern)  # save the relation to that CaseId
            else:  # if the relation is "illegal" according to B
                self.__inc[case_id] += 1  # increment incorrect for that caseID

            # Step 2: compute online conformance values
            self.__conformance[case_id] = len(self.__obs[case_id]) / (
                    len(self.__obs[case_id]) + self.__inc[case_id])  # calculated conformance

            if new_pattern in self.__B:  # if the relation is legal in B
                # if the relation occurrence is within P_min and P_max
                if self.__P[new_pattern][0] <= len(self.__obs[case_id]) and \
                        len(self.__obs[case_id]) <= self.__P[new_pattern][1]:
                    self.__completeness[case_id] = 1  # set completeness for that caseID
                else:  # if not within P_min and P_max
                    self.__completeness[case_id] = min(1, len(self.__obs[case_id]) / (
                            self.__P[new_pattern][0] + 1))  # calculate completeness

                self.__confidence[case_id] = 1 - (self.__F[new_pattern] / self.__maxOfMinRelationsAfter)
            else:
                self.__confidence[case_id] = -1
                self.__completeness[case_id] = -1

            # Step 3: cleanup
            self.__trace_last_event[case_id] = event_name

    def get_conformance(self, case_id=None):
        if case_id is None:
            return self.__conformance
        return self.__conformance[case_id]

    def get_completeness(self, case_id=None):
        if case_id is None:
            return self.__completeness
        return self.__completeness[case_id]

    def get_confidence(self, case_id=None):
        if case_id is None:
            return self.__confidence
        return self.__confidence[case_id]

    def get_model(self):
        return (self.__B, self.__P, self.__F)

    def set_model(self, M):
        self.__B = M[0]
        self.__P = M[1]
        self.__F = M[2]
        self.__maxOfMinRelationsAfter = 0
        for relations in self.__F:
            self.__maxOfMinRelationsAfter = max(self.__F[relations], self.__maxOfMinRelationsAfter)

    # def getMaxOfF(self):
    #     return self.__maxOfMinRelationsAfter
    #
    # def setM_FromXES(self, fileName):
    #     self.setM(model_from_XES().set_model_from_XES(fileName))

    # def printRefModel(self):
    #     G = pgv.Graph(comment='ref_model')  # setting type of graph
    #     G.graph_attr['rankdir'] = 'LR'
    #     G.node_attr['shape'] = 'box'
    #     G.edge_attr['arrowType'] = 'normal'
    #
    #     for (A, B) in self.__B:
    #         G.edge(A, B, label=str(), dir='forward')
    #
    #     G.render('model-output/ref_Model.gv').replace('\\', '/')


# Class originally developed by Magnus Frederiksen as part of his BSc project at DTU entitled
# "Development of Process Mining and Complex Event Processing using Python"
class BehavioralModelBuilder():
    def __init__(self):
        self.__B = []
        self.__P = dict()
        self.__F = dict()
        self.__tracelogs = defaultdict(list)
        self.__mark = []  # used to find P_max for all relation
        self.__mark2 = []

    def get_model(self):
        return self.__B, self.__P, self.__F

    def ingest_event(self, newEvent):
        case_id = newEvent.get_trace_name()  # locally save caseID
        event_name = newEvent.get_event_name()  # locally save event
        if case_id in self.__tracelogs:  # if this caseID has occured before
            relation = (self.__tracelogs[case_id][-1], event_name)  # find relation based on former event
            traceLength = len(self.__tracelogs[case_id]) - 1  # find the length of the trace so far
            if relation not in self.__B:  # if this relation already exist
                self.__B.append(relation)  # add relation

        else:  # if the trace hasent been seen yet
            self.__tracelogs[case_id] = []

        self.__tracelogs[case_id].append(event_name)  # add the event to the trace log

    def __setF(self, relation):
        Queue = []  # breath first search from an accepting state
        BFSMark = []
        BFSMark.append(relation)
        Queue.append((relation, 0))
        while len(Queue) > 0:
            ((A, B), depth) = Queue.pop(0)
            self.__F[(A, B)] = min(self.__F[(A, B)], depth)
            for (C, D) in self.__B:
                if D == A and (C, D) not in BFSMark:
                    BFSMark.append((C, D))
                    Queue.append(((C, D), depth + 1))

    def __findP_max(self, relation, depth):
        if relation not in self.__mark:  # brute force to find longest path
            self.__mark.append(relation)

            (P_min, P_max) = self.__P[relation]
            self.__P[relation] = (P_min, max(depth, P_max))

            (A, B) = relation
            for (C, D) in self.__B:
                if C == B and depth <= 5:
                    self.__findP_max((C, D), depth + 1)

            self.__mark.remove(relation)
        else:
            if relation in self.__mark2:
                return
            self.__mark2.append(relation)

            (P_min, P_max) = self.__P[relation]
            self.__P[relation] = (P_min, max(depth, P_max))
            (A, B) = relation
            for (C, D) in self.__B:
                if C == B and depth <= 5:
                    self.__findP_max((C, D), depth)

            self.__mark2.remove(relation)

    def __setP(self, relation):
        Queue = []  # breath first state from beggining state
        BFSMark = []
        BFSMark.append(relation)
        Queue.append((relation, 0))
        while len(Queue) > 0:
            ((A, B), depth) = Queue.pop(0)
            (P_min, P_max) = self.__P[(A, B)]
            self.__P[(A, B)] = (min(depth, P_min), max(depth, P_max))
            for (C, D) in self.__B:
                if C == B and (C, D) not in BFSMark:
                    BFSMark.append((C, D))
                    Queue.append(((C, D), depth + 1))

        self.__findP_max(relation, 0)

    def end_xes_to_model(self):
        for relation in self.__B:
            self.__F[relation] = sys.maxsize
            self.__P[relation] = (sys.maxsize, -1)

        endState = []
        startState = []

        for caseID, trace in self.__tracelogs.items():
            if trace[-1] not in endState:
                endState.append(trace[-1])
            if trace[0] not in startState:
                startState.append(trace[0])

        for (A, B) in self.__B:
            if A in startState:
                self.__setP((A, B))

        for (A, B) in self.__B:
            if B in endState:
                self.__setF((A, B))



import json
from typing import Optional

import paho.mqtt.client as mqtt
from reactivex import Observable, abc

from pybeamline.bevent import BEvent


def mqttxes_source(
        broker: str,
        port: int,
        base_topic: str) -> Observable[BEvent]:

    def subscribe(
            observer: abc.ObserverBase[BEvent],
            scheduler_: Optional[abc.SchedulerBase] = None
    ) -> abc.DisposableBase:
        def on_connect(client, userdata, flags, rc):
            if rc == 0:
                print("Connected to MQTT broker")
                if base_topic.endswith("/"):
                    client.subscribe(base_topic + '#')
                else:
                    client.subscribe(base_topic + '/#')
            else:
                observer.on_error(f"Connection failed with code {rc}")

        def on_message(client, userdata, msg):
            path = msg.topic.replace(base_topic, '').split('/')
            if path[0] == '':
                path.pop(0)
            activity_name = path[len(path) - 1]
            case = path[len(path) - 2]
            process = "Process"
            if len(path) > 2:
                process = path[len(path) - 3]
            e = BEvent(activity_name, case, process)
            try:
                if msg.payload.decode().strip() != '':
                    attributes = json.loads(msg.payload.decode())
                    for k in attributes:
                        e.event_attributes[k] = attributes[k]
            except json.JSONDecodeError:
                print("Error decoding JSON")
            observer.on_next(e)

        def on_disconnect(client, userdata, rc):
            if rc != 0:
                observer.on_error(f"Unexpected disconnection. Code: {rc}")
            else:
                observer.on_completed()

        client = mqtt.Client()
        client.on_connect = on_connect
        client.on_message = on_message
        client.on_disconnect = on_disconnect

        client.connect(broker, port, 60)
        client.loop_start()

        def dispose():
            client.loop_stop()
            client.disconnect()

        return dispose

    return Observable(subscribe)



from pybeamline.bevent import BEvent
from typing import Optional, Union
from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable
from pm4py.objects.log.obj import EventLog
from pm4py.util import xes_constants as xes_util
from pm4py import read_xes, convert_to_dataframe
import pandas as pd


def xes_log_source_from_file(log: str) -> Observable[BEvent]:
    return xes_log_source(read_xes(log))


def xes_log_source(
        log: Union[EventLog, pd.DataFrame], scheduler: Optional[abc.SchedulerBase] = None
) -> Observable[BEvent]:
    if type(log) is not pd.DataFrame:
        log = convert_to_dataframe(log)
    if xes_util.DEFAULT_TIMESTAMP_KEY in log.columns:
        log = log.sort_values(by=[xes_util.DEFAULT_TIMESTAMP_KEY])

    def subscribe(
            observer: abc.ObserverBase[BEvent], scheduler_: Optional[abc.SchedulerBase] = None
    ) -> abc.DisposableBase:

        for index, event in log.iterrows():
            time = None
            if xes_util.DEFAULT_TIMESTAMP_KEY in event:
                time = event[xes_util.DEFAULT_TIMESTAMP_KEY]
            e = BEvent(
                event[xes_util.DEFAULT_NAME_KEY],
                event["case:" + xes_util.DEFAULT_TRACEID_KEY],
                "log-file",
                time)
            for col in log.columns:
                if col not in [xes_util.DEFAULT_NAME_KEY, "case:" + xes_util.DEFAULT_NAME_KEY, xes_util.DEFAULT_TIMESTAMP_KEY]:
                    if event[col] == event[col]: # verify for nan
                        if col.startswith("case:"):
                            e.trace_attributes[col[5:]] = event[col]
                        else:
                            e.event_attributes[col] = event[col]
            observer.on_next(e)
        observer.on_completed()
        return CompositeDisposable()

    return Observable(subscribe)



def dfg_to_graphviz(dfg):
    graphviz = """
    digraph G {

    ranksep = 0.5
    fontsize = 9
    remincross = true
    margin = "0.0,0.0"
    outputorder = "edgesfirst"

    node[
        shape = box
        height = 0.23
        width = 1.2
        style = "rounded,filled"
        fontname = "Arial"
    ]
    edge[
        decorate = false
        fontsize = 8
        arrowsize = 0.5
        fontname = Arial
        tailclip = false
    ]
    
    start [
        shape = circle
        style = filled
        fillcolor = "#CED6BD"
        gradientangle = 270
        color = "#595F45"
        height = 0.13
        width = 0.13
        label = ""
    ]
    end [
        shape = circle
        style = filled
        fillcolor = "#D8BBB9"
        gradientangle = 270
        color = "#614847"
        height = 0.13
        width = 0.13
        label = ""
    ]
    """
    # Add all regular edges
    for (source, target), weight in dfg.items():
        graphviz += f'"{source}" -> "{target}" [penwidth={ max(0.5, min(5 * weight, 5)) },label="{ round(weight, 2) }"];\n'

    # Add edges to start and end nodes for all nodes that have no incoming or outgoing edges
    start_nodes = {source for source, target in dfg.keys()}
    end_nodes = {target for source, target in dfg.keys()}
    for (source, target) in dfg.keys():
        if source != target:
            if source in end_nodes:
                end_nodes.remove(source)
            if target in start_nodes:
                start_nodes.remove(target)
    for node in start_nodes:
        graphviz += f'start -> "{node}" [penwidth = 2, style = dashed, color = "#ACB89C"];\n'
    for node in end_nodes:
        graphviz += f'"{node}" -> end [penwidth = 2, style = dashed, color = "#C2B0AB"];\n'

    graphviz += "}"
    return graphviz
